#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 22:55:02 2019

@author: carl
"""

import traceback
import os.path
from collections import namedtuple
import numpy as np
import scipy.signal, scipy.io
from scipy.interpolate import interp1d
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot

from octavvs.io import SpectralData, Parameters
from octavvs.algorithms import baseline, correction, normalization, ptir

class PrepParameters(Parameters):
    """
    A class representing all the settings that can be made in the
    preprocessing UI and used to start a batch job.
    """

    def setAlgorithm(self, algo: str):
        algs = ['bassan', 'konevskikh', 'rasskazov']
        alg = next((a for a in algs if a in algo.lower()), None)
        if alg is None:
            raise ValueError('Algorithm must be one of '+str(algs))
        self.scAlgorithm = alg

    def load(self, filename):
        super().load(filename)
        self.setAlgorithm(self.scAlgorithm)



class PrepWorker(QObject):
    """
    Worker thread class for the heavy parts of the preprocessing: Scattering
    correction and the multiple-file batch processing.
    """
    # Signals for when processing is finished or failed, and for progress indication
    done = pyqtSignal(np.ndarray, np.ndarray, np.ndarray)
    stopped = pyqtSignal()
    failed = pyqtSignal(str, str)
    progress = pyqtSignal(int, int)
    progressPlot = pyqtSignal(np.ndarray, tuple)

    fileLoaded = pyqtSignal(int)
    loadFailed = pyqtSignal(str, str, str)

    batchProgress = pyqtSignal(int, int)
    batchDone = pyqtSignal(bool)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.halt = False

    @pyqtSlot(int)
    def loadFile(self, data, num):
        "Load file number num in the data object, emitting a signal on failure"
        try:
            file = data.filenames[num]
            if file == data.curFile:
                return True
            data.read_matrix(file)
        except (RuntimeError, FileNotFoundError) as e:
            self.loadFailed.emit(file, str(e), '')
        except Exception as e:
            self.loadFailed.emit(file, str(e), traceback.format_exc())
        else:
            self.fileLoaded.emit(num)
            return True
        return False

    def emitProgress(self, *pargs):
        "Combined progress signal and check for user interruption"
        if self.halt:
            raise InterruptedError('interrupted by user')
        self.progress.emit(*pargs)

    def loadReference(self, data, ref, otherref='', percentile=50):
        """
        Load an RMieSC reference spectrum, at wavenumbers that match the
        currently loaded file.
        """
        if ref == 'Mean':
            return data.raw.mean(0)
        elif ref == 'Percentile':
            return np.percentile(data.raw, percentile, axis=0)
        elif ref == 'Other':
            if otherref == '':
                raise RuntimeError('Specify a reference spectrum file')
            return correction.load_reference(data.wavenumber,
                                             matfilename=otherref)
        else:
            return correction.load_reference(data.wavenumber, what=ref.lower())

    def callACandSC(self, data, params, wn, y):
        """
        Run atmospheric correction and/or CRMieSC and/or mIRage correction
        """
        if params.mcDo:
            self.emitProgress(-5, 100)
            y = ptir.normalize_mirage(wn, y,
                                      endpoints=params.mcEndpoints,
                                      slopefactor=params.mcSlopefactor)[0]

        if params.acDo:
            self.emitProgress(-1, 100)
            y = correction.atmospheric(wn, y, cut_co2=params.acSpline,
                                       extra_iters=5 if params.acLocal else 0,
                                       smooth_win=9 if params.acSmooth else 0,
                                       atm=params.acReference,
                                       progressCallback=self.emitProgress)[0]

        if params.scDo:
            self.emitProgress(-2, 100)
            ref = self.loadReference(data, params.scRef,
                                     otherref=params.scOtherRef,
                                     percentile=params.scRefPercentile)
            yold = y
            clust = params.scClusters * (-1 if params.scStable else 1) if \
                params.scClustering else 0
#            print(params.scAmin, params.scAmax, params.scResolution)
            modelparams=dict(
                n_components=params.scPCAMax if \
                    params.scPCADynamic else params.scPCA,
                variancelimit=params.scPCAVariance*.01 if \
                    params.scPCADynamic else 0,
                a=np.linspace(params.scAmin, params.scAmax, params.scResolution),
                d=np.linspace(params.scDmin, params.scDmax, params.scResolution),
                bvals=params.scResolution,
                model=params.scAlgorithm,
                constantcomponent=params.scConstant,
                linearcomponent=params.scLinear)
            y = correction.rmiesc(
                    wn, y, ref,
                    iterations=params.scIters,
                    clusters=clust,
                    modelparams=modelparams,
                    weighted=False,
                    autoiterations=params.scAutoIters,
                    targetrelresiduals=1-params.scMinImprov*.01,
                    zeroregionpenalty=params.scPenalizeLambda if
                    params.scPenalize else None,
                    prefit_reference=params.scPrefitReference,
                    verbose=True,
                    progressCallback=self.emitProgress,
                    progressPlotCallback=self.progressPlot.emit)
            self.done.emit(wn, yold, y)

        return y

    def callSGFandSRandBC(self, params, wn, y):
        """
        Run three more preprocessing steps
        """
        if params.sgfDo:
            self.emitProgress(-3, 100)
            y = scipy.signal.savgol_filter(
                y, params.sgfWindow, params.sgfOrder, axis=1)

        if params.srDo:
            a = len(wn) - wn[::-1].searchsorted(params.srMax, 'right')
            b = len(wn) - wn[::-1].searchsorted(params.srMin, 'left')
            wn = wn[a:b]
            y = y[:, a:b]

        if params.bcDo:
            self.emitProgress(-4, 100)
            if params.bcMethod == 'rubberband':
                y -= baseline.rubberband(
                        wn, y, progressCallback=self.emitProgress)
            elif params.bcMethod == 'concaverubberband':
                y -= baseline.concaverubberband(
                        wn, y, iters=params.bcIters,
                        progressCallback=self.emitProgress)
            elif params.bcMethod == 'asls':
                y -= baseline.asls(
                        y, lam=params.bcLambda, p=params.bcP,
                        progressCallback=self.emitProgress)
            elif params.bcMethod == 'arpls':
                y -= baseline.arpls(
                        y, lam=params.bcLambdaArpls,
                        progressCallback=self.emitProgress)
            else:
                raise ValueError('unknown baseline correction method '+str(params.bcMethod))

        return wn, y

    @pyqtSlot(SpectralData, dict)
    def rmiesc(self, data, params):
        """ Run RMieSC, possibly preceded by atmospheric correction, on
        all or a subset of the raw data.
        Parameters:
            data: SpectralData object with raw data
            params: dictionary, mostly with things from PrepParameters (see code)
        """
        try:
            params['scDo'] = True
            params = namedtuple('rmiescParams', params.keys())(*params.values())
            if params.selected is not None:
                y = data.raw[params.selected]
            else:
                y = data.raw
            self.callACandSC(data, params, data.wavenumber, y)

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())


    def saveCorrected(self, outfile, fmt, data, wn, y):
        if fmt == 'Quasar.mat':
            out = {'y': y, 'wavenumber': wn}
            if data.pixelxy is not None:
                map_x = np.array([x for (x,y) in data.pixelxy])
                map_y = np.array([y for (x,y) in data.pixelxy])
            else:
                map_x = np.tile(data.wh[0], data.wh[1])
                map_y = np.repeat(range(data.wh[0]), data.wh[1])
            out['map_x'] = map_x[:, None]
            out['map_y'] = map_y[:, None]
            scipy.io.savemat(outfile, out)
        else:
            ab = np.hstack((wn[:, None], y.T))
            scipy.io.savemat(outfile, {'AB': ab, 'wh': data.wh } )

    @pyqtSlot(SpectralData, PrepParameters, str, bool)
    def bigBatch(self, data, params, folder, preservepath):
        """
        Run the batch processing of all the files listed in 'data'
        Parameters:
            data: SpectralData object with one or more files
            params: PrepParameters object from the user
            folder: output directory
            preservepath: if True, all processed files whose paths are under
            data.foldername will be placed in the corresponding subdirectory
            of the output directory.
        """
        try:
            for fi in range(len(data.filenames)):
                self.batchProgress.emit(fi, len(data.filenames))
                if not self.loadFile(data, fi):
                    continue

                wn = data.wavenumber
                y = data.raw

                y = self.callACandSC(data, params, wn, y)
                wn, y = self.callSGFandSRandBC(params, wn, y)

                if params.normDo:
                    y = normalization.normalize_spectra(
                        params.normMethod, y, wn, wavenum=params.normWavenum)

                # Figure out where to save the file
                filename = data.curFile
                if preservepath and filename.startswith(data.foldername):
                    filename = filename[len(data.foldername):]
                    filename = folder + filename
                    os.makedirs(os.path.dirname(filename), exist_ok=True)
                else:
                    filename = os.path.join(folder, os.path.basename(filename))
                # Add the extension
                filename = os.path.splitext(filename)
#                filename = filename[0] + params.saveExt + filename[1]
                filename = filename[0] + params.saveExt + '.mat'

                self.saveCorrected(filename, params.saveFormat, data, wn, y)
            self.batchDone.emit(True)
            return

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())
        self.batchDone.emit(False)


    @pyqtSlot(SpectralData, PrepParameters, str)
    def createReference(self, data, params, outfile):
        """
        Create reference spectrum from all the files listed in 'data', using
        a subset of the processing steps.
        Parameters:
            data: SpectralData object with one or more files
            params: PrepParameters object from the user
            outfile: name of output file
        """
        params.scDo = False
        params.srDo = False
        wns = []
        ys = []
        try:
            for fi in range(len(data.filenames)):
                self.batchProgress.emit(fi, len(data.filenames))
                if not self.loadFile(data, fi):
                    continue

                wn = data.wavenumber
                y = data.raw

                if params.scRef == 'Percentile':
                    y = np.percentile(y, params.scRefPercentile, axis=0)[None, :]
                else:
                    y = y.mean(0)[None, :]

                y = self.callACandSC(data, params, wn, y)
                wn, y = self.callSGFandSRandBC(params, wn, y)
                if params.normDo:
                    y = normalization.normalize_spectra(
                        params.normMethod, y, wn,
                        wavenum=params.normWavenum)
                wns.append(wn)
                ys.append(y[0])

            # Do all images have the same wavenumbers?
            if all(np.array_equal(v, wns[0]) for v in wns):
                ys = np.median(ys, axis=0)
            else:
                w1 = min(v.min() for v in wns)
                w2 = max(v.max() for v in wns)
                maxres = max((len(v) - 1) / (v.max() - v.min()) for v in wns)
                wn = np.linspace(w1, w2, num=maxres * (w2 - w1) + 1)
                interpol = interp1d(np.concatenate(wns), np.concatenate(ys))
                ys = interpol(wn)

            ab = np.hstack((wn[:, None], ys[:, None]))
            scipy.io.savemat(outfile, {'AB': ab } )
            self.batchDone.emit(True)
            return

        except InterruptedError:
            self.stopped.emit()
        except Exception as e:
            traceback.print_exc()
            self.failed.emit(repr(e), traceback.format_exc())
        self.batchDone.emit(False)


class ABCWorker(QObject):
    """
    A smaller worker thread class for atmospheric and baseline correction only.
    """
    acDone = pyqtSignal(np.ndarray, np.ndarray, np.ndarray, np.ndarray)
    acFailed = pyqtSignal(str)
    bcDone = pyqtSignal(np.ndarray, np.ndarray, np.ndarray)
    bcFailed = pyqtSignal(str)

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.haltBC = False

    @pyqtSlot(np.ndarray, np.ndarray, dict)
    def ac(self, wn, y, params):
        """
        Run baseline correction, emitting the processed data
        """
        try:
            corr, factors = correction.atmospheric(
                wn, y, cut_co2=params['cut_co2'],
                extra_iters=5 if params['extra'] else 0,
                smooth_win=9 if params['smooth'] else 0,
                atm=params['ref'])
            self.acDone.emit(wn, y, corr, factors)
        except Exception:
            self.acFailed.emit(traceback.format_exc())

    def checkHaltBC(self, a, b):
        if self.haltBC:
            raise InterruptedError('interrupted by user')

    @pyqtSlot(np.ndarray, np.ndarray, str, dict)
    def bc(self, wn, y, method, params):
        try:
            self.checkHaltBC(0, 1)
            if method in {'rubberband', 'concaverubberband'}:
                corr = getattr(baseline, method)(wn, y, **params,
                              progressCallback=self.checkHaltBC)
            else:
                corr = getattr(baseline, method)(y, **params,
                              progressCallback=self.checkHaltBC)
            self.bcDone.emit(wn, y, corr)
        except InterruptedError:
            self.bcFailed.emit('')
        except Exception:
            self.bcFailed.emit(traceback.format_exc())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 14:46:53 2020

"""

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel
from PyQt5.QtGui import QPainter, QFontMetrics


class ElidedLabel(QLabel):
    "A QLabel... that should be used with horizontal policy = ignore"
    def paintEvent(self, event):
        painter = QPainter(self)
        metrics = QFontMetrics(self.font())
        elided = metrics.elidedText(self.text(), Qt.ElideRight, self.width())
        painter.drawText(self.rect(), self.alignment(), elided)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 19:59:34 2020

@author: carl
"""

import os
import os.path
import fnmatch
import traceback
from pkg_resources import resource_filename

from PyQt5.QtWidgets import QWidget, QInputDialog
from PyQt5.QtWidgets import QMessageBox
#from PyQt5.QtCore import QThread, pyqtSignal, pyqtSlot, Qt
from PyQt5 import uic

from . import NoRepeatStyle

FileLoaderUi = uic.loadUiType(resource_filename(__name__, "fileloader.ui"))[0]

class FileLoaderWidget(QWidget, FileLoaderUi):
    """
    Widget with a bunch of graphical components related to loading a list of files
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.spinBoxFileNumber.setStyle(NoRepeatStyle())
        self.comboBoxSingMult.currentIndexChanged.connect(
            lambda ix: self.lineEditLoadFilter.setEnabled(ix > 0))
        self.pushButtonFilePrev.clicked.connect(
            lambda: self.spinBoxFileNumber.setValue(
                self.spinBoxFileNumber.value() - 1))
        self.pushButtonFileNext.clicked.connect(
            lambda: self.spinBoxFileNumber.setValue(
                self.spinBoxFileNumber.value() + 1))

    def setEnabled(self, onoff):
        self.pushButtonLoad.setEnabled(onoff)
        self.pushButtonAdd.setEnabled(onoff)
        self.spinBoxFileNumber.setEnabled(onoff)
        self.pushButtonShowFiles.setEnabled(onoff)
        self.pushButtonFilePrev.setEnabled(onoff)
        self.pushButtonFileNext.setEnabled(onoff)

    def saveParameters(self, p):
        "Copy from UI to some kind of parameters object"
        p.fileFilter = self.lineEditLoadFilter.text()

    def loadParameters(self, p):
        "Copy to UI from some kind of parameters object"
        self.lineEditLoadFilter.setText(p.fileFilter)



class FileLoader():
    """
    Mixin class that adds file loading UI stuff to an OctavvsMainWindow.

    Things assumed to exist in self:
        fileLoader - FileLoader (from UI)
        settings - QSettings (from OctavvsApplication)
        data - SpectralData
    """

#    dimensionsUpdated = pyqtSignal(tuple)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fileLoader.pushButtonLoad.clicked.connect(self.loadFolder)
        self.fileLoader.pushButtonAdd.clicked.connect(self.addFolder)
        self.fileLoader.spinBoxFileNumber.valueChanged.connect(self.selectFile)
        self.fileLoader.pushButtonShowFiles.clicked.connect(self.showFileList)

        self.fileLoader.lineEditWidth.editingFinished.connect(
            lambda: self._updateDimension(0))
        self.fileLoader.lineEditHeight.editingFinished.connect(
            lambda: self._updateDimension(1))


    def addFolder(self):
        self.loadFolder(add=True)

    def loadFolder(self, add=False):
        mode = self.fileLoader.comboBoxSingMult.currentIndex()
        if mode > 0:
            recurse = mode == 2
            pattern = self.fileLoader.lineEditLoadFilter.text()

            foldername = self.getDirectoryName(
                    "Load spectra from directory (filter: %s)" % (pattern),
                    settingname='spectraDir')
            if not foldername:
                return

            filenames = []
            for root, dirs, files in os.walk(foldername):
                self.fileLoader.lineEditFilename.setText(root)
                if '/' in pattern:
                    files = [ os.path.join(root, f) for f in files ]
                    files = fnmatch.filter(files, pattern)
                    filenames += files
                else:
                    files = fnmatch.filter(files, pattern)
                    filenames += [ os.path.join(root, f) for f in files ]
                if not recurse:
                    break;
            if not filenames:
                self.errorMsg.showMessage('No matching files found in directory "' +
                                          foldername + '"')
                return
            filenames.sort()
        else:
            filenames = self.getLoadFileNames("Open hyperspectral image",
                        filter="Spectral data files (*.mat *.txt *.csv *.ptir *.0 *.1 *.2);;Matlab files (*.mat);;Text files (*.txt *.csv);;OPUS files (*.0 *.1 *.2 *.3 *.4 *.5 *.6 *.7 *.8 *.9);;mIRage files (*.ptir);;All files (*)",
                        settingname='spectraDir')
            if not filenames:
                return
            foldername = os.path.dirname(filenames[0])

        if self.updateFileList(filenames, False, add=add):
            self.data.foldername = foldername
            self.settings.setValue('whitelightDir', foldername)

    def loadFile(self, file):
        "Load a file or return (error message, traceback) from trying"
        try:
            self.data.read_matrix(file)
            return
        except (RuntimeError, FileNotFoundError) as e:
            return (str(e), None)
        except Exception as e:
            return (repr(e), traceback.format_exc())

    def updateFileListInfo(self, filenames):
        "Helper function for updateFileList"
        self.data.filenames = filenames
        self.fileLoader.labelFileCount.setText('of %d' % (len(filenames)))
        self.fileLoader.spinBoxFileNumber.setMaximum(len(filenames))
        self.fileLoader.spinBoxFileNumber.setEnabled(True)

    def updateFileList(self, filenames, avoidreload, add=False):
        """
        Updates the list of files and loads a file if possible
        Parameters:
            filenames: list of files from dialog/user
            avoidreload: if possible, select the current file in the list without reload
        Returns:
            True if a file was loaded, False if the user cancelled
        """
        if add:
            existing = set(self.data.filenames)
            filenames = self.data.filenames + [
                f for f in filenames if f not in existing]

        if avoidreload:
            # Is the current file already in our list?
            try:
                ix = filenames.index(self.data.curFile)
            except ValueError:
                pass
            else:
                self.updateFileListInfo(filenames)
                self.fileLoader.spinBoxFileNumber.setValue(ix+1)
                return True

        skipall = False
        while len(filenames):
            err = self.loadFile(filenames[0])
            if err is None:
                self.updateFileListInfo(filenames)
                self.updateFile(0)
                return True

            if not skipall:
                q = self.loadErrorBox(filenames[0], err)
                q.addButton('Abort', QMessageBox.RejectRole)
                if len(filenames) > 1:
                    q.addButton('Skip file', QMessageBox.AcceptRole)
                    q.addButton('Skip all with errors', QMessageBox.AcceptRole)
                ret = q.exec()
                if ret == 0:
                    break
                elif ret == 2:
                    skipall = True
            filenames = filenames[1:]
        return False

    def showFileList(self):
        s, ok = QInputDialog.getMultiLineText(
                self, 'Input files',
                'Editable list of input files',
                '\n'.join(self.data.filenames))
        if ok:
            fn = s.splitlines()
            if fn:
                self.updateFileList(fn, True)

    def selectFile(self):
        skipall = False
        while True:
            num = self.fileLoader.spinBoxFileNumber.value() - 1
            if (num < 0 or num >= len(self.data.filenames) or
                self.data.filenames[num] == self.data.curFile):
                return
            err = self.loadFile(self.data.filenames[num])
            if not err:
                self.updateFile(num)
                break
            elif not skipall:
                q = self.loadErrorBox(self.data.filenames[num], err)
                q.addButton('Skip file', QMessageBox.AcceptRole)
                q.addButton('Skip all with errors', QMessageBox.AcceptRole)
                ret = q.exec()
                if ret:
                    skipall = True

            del self.data.filenames[num]
            self.updateFileListInfo(self.data.filenames)

    def updateFile(self, num):
        file = self.data.filenames[num]
        self.fileLoader.spinBoxFileNumber.setValue(num+1)
        self.fileLoader.lineEditFilename.setText(file)
        self.fileLoader.lineEditWidth.setText(str(self.data.wh[0]))
        self.fileLoader.lineEditHeight.setText(str(self.data.wh[1]))

    def _updateDimension(self, dimnum):
        "Update width/height in data.wh"
        if dimnum == 0:
            self.data.set_width(self.fileLoader.lineEditWidth.text())
        else:
            self.data.set_height(self.fileLoader.lineEditHeight.text())
        self.fileLoader.lineEditWidth.setText(str(self.data.wh[0]))
        self.fileLoader.lineEditHeight.setText(str(self.data.wh[1]))
        self.updateDimensions(self.data.wh)

    def updateWavenumberRange(self):
        super().updateWavenumberRange()
        if self.data.wavenumber is not None:
            self.fileLoader.lineEditLength.setText(str(len(self.data.wavenumber)))





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 18 14:45:53 2021

@author: carl
"""

import numpy as np
import scipy
import sklearn
import sklearn.metrics
import time

def simplisma(d, nr, f):
    """
    The SIMPLISMA algorithm for finding a set of 'pure' spectra to serve
    as starting point for MCR-ALS etc.
    Reference Matlab Code:
        J. Jaumot, R. Gargallo, A. de Juan, R. Tauler,
        Chemometrics and Intelligent Laboratoty Systems, 76 (2005) 101-110

    Parameters
    ----------
    d : array(nspectra, nwavenums)
        input spectra.
    nr : int
        number of output components.
    f : float
        noise threshold.

    Returns
    -------
    spout: array(nr, nspectra)
        concentration profiles of 'purest' spectra.
    imp : array(nr, dtype=int)
        indexes of the 'purest' spectra.
    """

    nrow = d.shape[0]
    ncol = d.shape[1]
    s = d.std(axis=0)
    m = d.mean(axis=0)
    mf = m + m.max() * f
    p = s / mf

    # First Pure Spectral/Concentration profile
    imp = np.empty(nr, dtype=np.int)
    imp[0] = p.argmax()

    #Calculation of correlation matrix
    l2 = s**2 + mf**2
    dl = d / np.sqrt(l2)
    c = (dl.T @ dl) / nrow

    #calculation of the first weight
    w = (s**2 + m**2) / l2
    p *= w
    #calculation of following weights
    dm = np.zeros((nr+1, nr+1))
    for i in range(1, nr):
        dm[1:i+1,1:i+1] = c[imp[:i],:][:,imp[:i]]
        for j in range(ncol):
            dm[0,0] = c[j,j]
            dm[0,1:i+1]=c[j,imp[:i]]
            dm[1:i+1,0]=c[imp[:i],j]
            w[j] = np.linalg.det(dm[0:i+1, 0:i+1])

        imp[i] = (p * w).argmax()

    ss = d[:,imp]
    spout = ss / np.sqrt(np.sum(ss**2, axis=0))
    return spout.T, imp

# import pymcr


# def pymcr_als(sp, initial_components, maxiters, reltol,
#             callback_iter):

#     mcr = pymcr.mcr.McrAR(max_iter=maxiters,
#                           tol_err_change=reltol,
#                           tol_increase=1., tol_n_increase=10,
#                           tol_n_above_min=30)
#     mcr.fit(sp, ST=initial_components, post_iter_fcn=callback_iter)

#     return mcr.n_iter, mcr.C_opt_.T, mcr.ST_opt_, np.asarray(mcr.err)



def mcr_als(sp, initial_components, maxiters, nonnegative=(True, True),
            tol_rel_error=0, tol_ups_after_best=None, tol_time=None,
            callback=None, acceleration=None, **kwargs):
    # m=5,experimental=None, tol_time=None):
    """
    Perform MCR-ALS nonnegative matrix decomposition on the matrix sp

    Parameters
    ----------
    sp : array(nsamples, nfeatures)
        Spectra to be decomposed.
    initial_components : array(ncomponents, nfeatures)
        Initial concentrations.
    maxiters : int
        Maximum number of iterations.
    nonnegative : pair of bool
        True for the matrix/matrices that must be non-negative
    tol_rel_error : float, optional
        Error target (relative to first iteration).
    tol_ups_after_best : float, optional
        Stop after error going net up this many times since best error.
    tol_time : float, optional
        Stop after this many seconds of process time have elapsed
    callback : func(it : int, err : array(float), A : array, B : array)
        Callback for every half-iteration.
    acceleration : str
        None, 'Anderson', 'AdaptiveOverstep'.
    m : int, >1
        For Anderson acceleration: the number of earlier steps to consider.
    experimental : something
        Something

    Returns
    -------
    iterations : int
        Number of iterations performed
    A : array(ncomponents, nfeatures)
        Spectra (at lowest error)
    B : array(ncomponents, nsamples)
        Concentrations at lowest error
    error : list(float)
        Errors at all half-iterations
    """
    nrow, ncol = np.shape(sp)
    nr = initial_components.shape[0]
    # u, s, v = np.linalg.svd(sp)
    # s = scipy.linalg.diagsvd(s, nrow, ncol)
    # u = u[:, :nr]
    # s = s[:nr, :nr]
    # v = v[:nr, :]
    # dauxt = sklearn.preprocessing.normalize((u @ s @ v).T)
    # dauxt = sp.T
    A = initial_components.T.copy()
    B = np.empty((nr, nrow))
    errors = []
    errorbest = None # Avoid spurious warning

    prevA, prevB = (None, None)
    if acceleration == 'Anderson':
        ason_m = kwargs['m'] if 'm' in kwargs else 3
        g_ab = [None, None]
        G_ab = [[], []]
        X_ab = [[], []]
    elif acceleration == 'AdaptiveOverstep':
        if 'aosettings' in kwargs:
            ao = kwargs['aosettings']
            ao_stepsize = [ao[0], ao[0]]
            ao_factors = np.asarray(ao[1:5]).reshape((2, 2))
            ao_failscale = ao[5]
        else:
            ao_stepsize = [.5, .5]
            ao_factors = [[.92, -.02], [1.08, .02]]
            ao_failscale = .5
    elif acceleration:
        raise ValueError("acceleration must be None or 'Anderson'")

    if tol_time:
        endtime = time.process_time() + tol_time

    for it in range(maxiters * 2):
        update_A = it & 1

        if update_A:
            prevA = A
            if nonnegative[0]:
                A = np.empty_like(A)
                for i in range(ncol):
                    A[i, :] = scipy.optimize.nnls(B.T, sp[:, i])[0]
            else:
                A = np.linalg.solve(B.T, sp.T)
        else:
            prevB = B
            if nonnegative[1]:
                B = np.empty_like(B)
                for i in range(nrow):
                    B[:, i] = scipy.optimize.nnls(A, sp[i, :])[0]
            else:
                B = np.linalg.solve(A, sp).T

        # if acceleration == 'Anderson' and update_A:
        #     for ab in range(1):
        #         AB, prevAB = (B, prevB) if ab else (A, prevA)
        #         G = G_ab[ab]
        #         X = X_ab[ab]
        #         prevg = g_ab[ab]
        #         g_ab[ab] = (AB - prevAB).flatten()
        #         g = g_ab[ab]
        #         if len(X) < 1:
        #             X.append(g) # f(x0) - x0
        #         else:
        #             G.append(g - prevg)
        #             while(len(G) > ason_m):
        #                 G.pop(0)
        #                 X.pop(0)
        #             Garr = np.asarray(G)
        #             gamma = np.linalg.lstsq(Garr.T, g, rcond=-1)[0]
        #             print('gamma', ab, gamma, np.linalg.norm(g))
        #             xstep = g - gamma @ (np.asarray(X) + Garr)
        #             np.add(prevAB, xstep.reshape(AB.shape), out=AB)
        #             X.append(xstep)

        if acceleration == 'Anderson' and update_A:
            for ab in range(1):
                AB, prevAB = (B, prevB) if ab else (A, prevA)
                G = G_ab[ab]
                X = X_ab[ab]
                prevg = g_ab[ab]
                g_ab[ab] = (AB - prevAB).flatten()
                g = g_ab[ab]
                if len(X) < 1:
                    X.append(g) # f(x0) - x0
                else:
                    G.append(g - prevg)
                    while(len(G) > ason_m):
                        G.pop(0)
                        X.pop(0)
                    Garr = np.asarray(G)
                    gamma = np.linalg.lstsq(Garr.T, g, rcond=-1)[0]
                    print('gamma', ab, gamma, np.linalg.norm(g))
                    xstep = g - gamma @ (np.asarray(X) + Garr)
                    np.add(prevAB, xstep.reshape(AB.shape), out=AB)
                    X.append(xstep)

        if acceleration == 'AdaptiveOverstep' and it >= 4:
            steps = [[max(ao_stepsize[ab] * fac[0] + fac[1], 0)
                   for ab in range(2)] for fac in ao_factors]
            AB = []
            errorv = []
            for ss in steps:
                AB.append([A + ss[0] * (A - prevA), B + ss[1] * (B - prevB)])
                for nn in range(2):
                    if nonnegative[nn]:
                        AB[-1][nn] = np.maximum(0, AB[-1][nn])
                errorv.append(sklearn.metrics.mean_squared_error(
                    sp.T, np.dot(AB[-1][0], AB[-1][1])))
            mm = np.argmin(errorv)
            if errorv[mm] < errors[-1]:
                if update_A:
                    A = AB[mm][0]
                else:
                    B = AB[mm][1]
                error = errorv[mm]
                ao_stepsize = steps[mm]
            else:
                ao_stepsize = [s * ao_failscale for s in steps[mm]]
                error = sklearn.metrics.mean_squared_error(sp.T, np.dot(A, B))
        else:
            error = sklearn.metrics.mean_squared_error(sp.T, np.dot(A, B))

        errors.append(error)
        if not it or error < errorbest:
            errorbest = error
            Abest = A
            Bbest = B
            netups = 0
        elif it:
            if tol_ups_after_best is not None:
                if error < errors[-2]:
                    netups = max(0, netups - 1)
                else:
                    netups = netups + 1
                    if netups > tol_ups_after_best:
                        break
            if np.abs(error - errors[0]) / error < tol_rel_error:
                break
        if it and tol_time:
            if time.process_time() > endtime:
                break
        if callback is not None:
            callback(it, errors, A.T, B)

    return it // 2, Abest.T, Bbest, errors




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  4 02:56:27 2021

@author: carl
"""

import numpy as np

from .util import find_wn_ranges



def normalize_mirage(wn, data, breaks=[(922,926), (1202,1206), (1448,1452)],
                     endpoints=6, slopefactor=.5):
    """
    Correct for drift in quantum cascade laser intensity between the four
    regions of mIRage PTIR spectra.

    Near the end of each of the regions, spectra are be assumed to have the
    same level except for an overall per-spectrum intensity level. The
    normalization factors at the region ends are linearly interpolated
    within each of the regions, then applied. Finally, values in the
    break regions are interpolated anew from the surrounding points.

    Parameters
    ----------
    wn : array
        wavenumbers, sorted
    data : array(pixel, wn)
        raw mIRage data
    breaks : list of 2-ples, optional
        regions where the spectra are interpolated between the QCL chips
    endpoints: int
        number of wavenumbers to consider when normalizing levels at
        the edge of each region.

    Returns
    -------
    data : array
        corrected data
    scale : array(pixels, 8)
        rescaling factor for the spectra at start and end of each of the
        four regions defined by the breaks.

    """
    data = data.copy()
    if not len(data):
        return data, None
    flipwn = wn[0] > wn[-1]
    if flipwn:
        data = data[:, ::-1]
        wn = wn[::-1]
    breaks = find_wn_ranges(wn, np.array(breaks))
    cuts = np.concatenate(([0], breaks.flatten(), [len(wn)])).reshape((-1, 2))

    # Straight baseline of each segment. shape=(ncuts, 2, nspectra)
    scale = np.zeros(cuts.shape + (len(data),))
    # Overall level in each segment
    dsum = np.zeros(len(cuts))
    slopes = np.zeros(cuts.shape)
    cutpoints = np.zeros(cuts.shape)

    def linreg(x, y):
        xm = x.mean()
        ym = y.mean(1)
        rx = x - xm
        s = (rx * (y.T - ym).T).sum(1) / (rx * rx).sum()
        return xm, ym, s

    for i in range(len(cuts)):
        cb = cuts[i][0]
        ce = cuts[i][1]
        cwidth = min(endpoints, (ce - cb) // 2)

        wb = linreg(wn[cb:cb+cwidth], np.abs(data[:, cb:cb+cwidth]))
        we = linreg(wn[ce-cwidth:ce], np.abs(data[:, ce-cwidth:ce]))

        cutpoints[i, :] = [wb[0], we[0]]
        # sc = np.maximum([wb[1], we[1]], 1)
        sc = [wb[1], we[1]]
        scale[i,:,:] = sc
        slopes[i, :] = np.array([wb[2], we[2]]).mean(1)
        # need to handle negative values here!
        # dsum[i] = np.abs(data[:, cb:ce]).sum()
        dsum[i] = np.maximum(data[:, cb:ce], 0).sum()
    # Mean level of all spectra in each of the cut points
    means = scale.mean(2)
    # Make the mean levels identical on both sides of the cuts
    for i in range(len(means)-1):
        mm = min(means[i][1], means[i+1][0])
        ds = (slopes[i+1, 0] + slopes[i, 1]) / mm * (
            wn[cuts[i+1, 0]] - wn[cuts[i, 1]]) * slopefactor
        ds = max(-.5, min(.5, ds))
        means[i][1] = mm * (1 - ds)
        means[i+1][0] = mm * (1 + ds)
    # print('means', means)
    scale = (scale.T / means.T).T
    weights = dsum / dsum.mean()
    scale = scale / ((scale.min(1).T * weights).mean(1))
    # scale = scale / ((scale.mean(1).T * weights).mean(1))

    for i in range(len(cuts)):
        cb = cuts[i][0]
        ce = cuts[i][1]
        data[:, cb:ce] /= np.linspace(scale[i][0], scale[i][1], ce-cb, axis=-1)
        if i:
            pce = cuts[i-1][1]
            data[:, pce:cb] = np.linspace(data[:, pce-1], data[:, cb], cb-pce+1,
                                          endpoint=False, axis=-1)[:,1:]
    scale = scale.reshape((-1, len(data))).T
    if flipwn:
        data = data[:,::-1]
    return data, scale



import os
import traceback
# from os.path import basename, dirname
from pkg_resources import resource_filename
import argparse

from PyQt5.QtWidgets import QDialog, QMessageBox
from PyQt5.QtCore import QThread, pyqtSignal, pyqtSlot
from PyQt5 import uic

import numpy as np
import scipy.signal, scipy.io
import matplotlib
matplotlib.use('QT5Agg')
import matplotlib.pyplot as plt

from .prep.prepworker import PrepWorker, ABCWorker, PrepParameters
import octavvs.io
from octavvs.algorithms import normalization, ptir
from octavvs.io import SpectralData
from octavvs.ui import (FileLoader, ImageVisualizer, OctavvsMainWindow,
                        NoRepeatStyle, uitools)


Ui_MainWindow = uic.loadUiType(resource_filename(
    __name__, "prep/preprocessing_ui.ui"))[0]
Ui_DialogSCAdvanced = uic.loadUiType(resource_filename(
    __name__, "prep/scadvanced.ui"))[0]
Ui_DialogCreateReference = uic.loadUiType(resource_filename(
    __name__, "prep/create_reference.ui"))[0]


class DialogSCAdvanced(QDialog, Ui_DialogSCAdvanced):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)

class DialogCreateReference(QDialog, Ui_DialogCreateReference):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)

class MyMainWindow(ImageVisualizer, FileLoader, OctavvsMainWindow, Ui_MainWindow):

    closing = pyqtSignal()
    startRmiesc = pyqtSignal(SpectralData, dict)
    startAC = pyqtSignal(np.ndarray, np.ndarray, dict)
    startBC = pyqtSignal(np.ndarray, np.ndarray, str, dict)
    startBatch = pyqtSignal(SpectralData, PrepParameters, str, bool)
    startCreateReference = pyqtSignal(SpectralData, PrepParameters, str)

    bcLambdaRange = np.array([0, 8])
    bcPRange = np.array([-5, 0])

    @classmethod
    def program_name(cls):
        "Return the name of the program that this main window represents"
        return 'Preprocessing'

    def __init__(self, parent=None, files=None, paramFile=None, savePath=None):
        super().__init__(parent)

        # self.splitter.setSizes([1e5]*3)

        self.data = SpectralData()
        self.ptirMode = False
        self.rmiescRunning = 0   # 1 for rmiesc, 2 for batch, 3 for reference

        # Avoid repeating spinboxes
        self.spinBoxItersBC.setStyle(NoRepeatStyle())
        self.spinBoxNclusScat.setStyle(NoRepeatStyle())
        self.spinBoxNIteration.setStyle(NoRepeatStyle())
        self.spinBoxPolyOrder.setStyle(NoRepeatStyle())
        self.spinBoxWindowLength.setStyle(NoRepeatStyle())

        self.lineEditSaveExt.setText('_prep')

        self.imageVisualizer.plot_spectra.updated.connect(self.updateMC)
        self.plot_MC.clicked.connect(self.plot_MC.popOut)
        self.checkBoxMC.toggled.connect(self.updateMC)
        self.spinBoxMCEndpoints.valueChanged.connect(self.updateMC)
        self.lineEditMCSlopefactor.editingFinished.connect(self.updateMC)

        self.plot_MC.updated.connect(self.updateAC)
        self.plot_AC.clicked.connect(self.plot_AC.popOut)
        self.checkBoxAC.toggled.connect(self.updateAC)
        self.checkBoxSpline.toggled.connect(self.updateAC)
        self.checkBoxLocalPeak.toggled.connect(self.updateAC)
        self.checkBoxSmoothCorrected.toggled.connect(self.updateAC)
        self.pushButtonACLoadReference.clicked.connect(self.loadACReference)
        self.lineEditACReference.editingFinished.connect(self.updateAC)

        self.plot_AC.updated.connect(self.updateSC)
        self.plot_SC.clicked.connect(self.plot_SC.popOut)
        self.imageVisualizer.plot_raw.changedSelected.connect(self.updateSCplot)
        self.pushButtonSCRefresh.clicked.connect(self.refreshSC)
        self.comboBoxReference.currentIndexChanged.connect(self.updateSC)
        self.pushButtonLoadOther.clicked.connect(self.loadOtherReference)
        self.comboBoxReference.currentIndexChanged.connect(self.updateSC)
        self.lineEditSCRefPercentile.editingFinished.connect(self.updateSC)
        self.pushButtonLoadOther.clicked.connect(self.updateSC)
        self.spinBoxNIteration.valueChanged.connect(self.updateSC)
        self.checkBoxClusters.toggled.connect(self.scClustersToggle)
        self.spinBoxNclusScat.valueChanged.connect(self.updateSC)
        self.checkBoxStabilize.toggled.connect(self.updateSC)
        self.pushButtonSCStop.clicked.connect(self.scStop)
        self.pushButtonRunStop.clicked.connect(self.scStop)

        self.dialogCreateReference = DialogCreateReference()
        self.pushButtonCreateReference.clicked.connect(self.dialogCreateReference.show)
        self.pushButtonCreateReference.clicked.connect(self.dialogCreateReference.raise_)
        self.dialogCreateReference.pushButtonCreateReference.clicked.connect(self.createReference)
        self.dialogCreateReference.pushButtonStop.clicked.connect(self.scStop)

        self.dialogSCAdvanced = DialogSCAdvanced()
        self.pushButtonSCAdvanced.clicked.connect(self.dialogSCAdvanced.show)
        self.pushButtonSCAdvanced.clicked.connect(self.dialogSCAdvanced.raise_)
        self.dialogSCAdvanced.comboBoxSCAlgo.currentIndexChanged.connect(self.updateSC)
        self.dialogSCAdvanced.spinBoxSCResolution.valueChanged.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCamin.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCamax.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCdmin.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCdmax.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.checkBoxSCConstant.toggled.connect(self.updateSC)
        self.dialogSCAdvanced.checkBoxSCLinear.toggled.connect(self.updateSC)
        self.dialogSCAdvanced.checkBoxSCPrefitReference.toggled.connect(self.updateSC)
        self.dialogSCAdvanced.checkBoxSCPenalize.toggled.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCPenalizeLambda.setRange(1e-3, 1e3)
        self.dialogSCAdvanced.lineEditSCPenalizeLambda.setFormat("%g")
        self.dialogSCAdvanced.lineEditSCPenalizeLambda.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.spinBoxSCPCA.valueChanged.connect(self.updateSC)
        self.dialogSCAdvanced.radioButtonSCPCAFixed.toggled.connect(self.toggleSCPCA)
        self.dialogSCAdvanced.radioButtonSCPCADynamic.toggled.connect(self.toggleSCPCA)
        self.dialogSCAdvanced.spinBoxSCPCAMax.valueChanged.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCPCAVariance.setFormat("%.8g")
        self.dialogSCAdvanced.lineEditSCPCAVariance.setRange(50, 99.999999)
        self.dialogSCAdvanced.lineEditSCPCAVariance.editingFinished.connect(self.updateSC)
        self.dialogSCAdvanced.checkBoxSCAutoIters.toggled.connect(self.updateSC)
        self.dialogSCAdvanced.lineEditSCMinImprov.setRange(0, 99)
        self.dialogSCAdvanced.lineEditSCMinImprov.setFormat("%g")
        self.dialogSCAdvanced.lineEditSCMinImprov.editingFinished.connect(self.updateSC)

        self.plot_AC.updated.connect(self.updateSGF)
        self.plot_SGF.clicked.connect(self.plot_SGF.popOut)
        self.checkBoxSC.toggled.connect(self.updateSGF)
        self.plot_SC.updated.connect(self.updateSGF)
        self.checkBoxSGF.toggled.connect(self.updateSGF)
        self.spinBoxWindowLength.valueChanged.connect(self.updateSGF)
        self.spinBoxPolyOrder.valueChanged.connect(self.updateSGF)

        self.plot_SGF.updated.connect(self.updateSR)
        self.plot_SR.clicked.connect(self.plot_SR.popOut)
        self.checkBoxSR.toggled.connect(self.updateSR)
        self.horizontalSliderMin.valueChanged.connect(self.srMinSlide)
        self.horizontalSliderMax.valueChanged.connect(self.srMaxSlide)
        self.lineEditMinwn.editingFinished.connect(self.srMinEdit)
        self.lineEditMaxwn.editingFinished.connect(self.srMaxEdit)

        self.plot_SR.updated.connect(self.updateBC)
        self.plot_BC.clicked.connect(self.plot_BC.popOut)
        self.comboBoxBaseline.currentIndexChanged.connect(self.bcMethod)
        self.checkBoxBC.toggled.connect(self.updateBC)
        self.connectLogSliders()
        self.spinBoxItersBC.valueChanged.connect(self.updateBC)

        self.plot_BC.updated.connect(self.updateNorm)
        self.plot_norm.clicked.connect(self.plot_norm.popOut)
        self.checkBoxNorm.toggled.connect(self.updateNorm)
        self.comboBoxNormMethod.currentIndexChanged.connect(self.updateNorm)
        self.lineEditNormWavenum.editingFinished.connect(self.updateNorm)

        self.pushButtonSaveParameters.clicked.connect(self.saveParameters)
        self.pushButtonLoadParameters.clicked.connect(self.loadParameters)

        self.pushButtonRun.clicked.connect(self.runBatch)

        # Defaults when no data loaded
        self.scSettings = {}

        self.worker = PrepWorker()
        self.workerThread = QThread()
        self.worker.moveToThread(self.workerThread)
        self.closing.connect(self.workerThread.quit)
        self.startRmiesc.connect(self.worker.rmiesc)
        self.worker.done.connect(self.scDone)
        self.worker.stopped.connect(self.scStopped)
        self.worker.failed.connect(self.scFailed)
        self.worker.progress.connect(self.scProgress)
        self.worker.batchProgress.connect(self.batchProgress)
        self.worker.progressPlot.connect(self.plot_SC.progressPlot)
        self.worker.fileLoaded.connect(self.updateFile)
        self.worker.loadFailed.connect(self.showLoadErrorMessage)
        self.startBatch.connect(self.worker.bigBatch)
        self.startCreateReference.connect(self.worker.createReference)
        self.worker.batchDone.connect(self.batchDone)
        self.workerThread.start()

        self.bcNext = None
        self.acNext = None

        self.abcWorker = ABCWorker()
        self.abcWorkerThread = QThread()
        self.abcWorker.moveToThread(self.abcWorkerThread)
        self.closing.connect(self.abcWorkerThread.quit)
        self.startBC.connect(self.abcWorker.bc)
        self.startAC.connect(self.abcWorker.ac)
        self.abcWorker.acDone.connect(self.acDone)
        self.abcWorker.bcDone.connect(self.bcDone)
        self.abcWorker.acFailed.connect(self.acFailed)
        self.abcWorker.bcFailed.connect(self.bcFailed)
        self.abcWorkerThread.start()

        self.lineEditMinwn.setFormat("%.2f")
        self.lineEditMaxwn.setFormat("%.2f")
        self.lineEditBCThresh.setRange(1e-6, 1e6)

        self.lineEditMCSlopefactor.setFormat("%g")
        self.lineEditMCSlopefactor.setRange(-1, 2)

        self.lineEditSCRefPercentile.setFormat("%g")
        self.lineEditSCRefPercentile.setRange(0., 100.)
        self.dialogSCAdvanced.lineEditSCamin.setFormat("%.4g")
        self.dialogSCAdvanced.lineEditSCamax.setFormat("%.4g")
        self.dialogSCAdvanced.lineEditSCdmin.setFormat("%.4g")
        self.dialogSCAdvanced.lineEditSCdmax.setFormat("%.4g")
        self.dialogSCAdvanced.lineEditSCamin.setRange(1., 4.)
        self.dialogSCAdvanced.lineEditSCamax.setRange(1., 4.)
        self.dialogSCAdvanced.lineEditSCdmin.setRange(.1, 100.)
        self.dialogSCAdvanced.lineEditSCdmax.setRange(.1, 100.)

        self.updateWavenumberRange()
        self.bcMethod()

        self.post_setup()
        if files is not None and files != []:
            self.updateFileList(files, False) # Load files passed as arguments

        if paramFile is not None: # Loads the parameter file passed as argument
            self.loadParameters(filename=paramFile)

        if savePath is not None:
            if paramFile is None:
                self.errorMsg.showMessage("Running from the command line "+
                                          "without passing a parameter file does nothing.")
            savePath = os.path.normpath(savePath)
            self.runBatch(foldername=savePath)

    def closeEvent(self, event):
        self.worker.halt = True
        self.abcWorker.haltBC = True
        self.closing.emit()
        plt.close('all')
        self.workerThread.wait()
        self.abcWorkerThread.wait()
        self.deleteLater()
        self.dialogSCAdvanced.close()
        self.dialogCreateReference.close()
#        qApp.quit()

    def updateWavenumberRange(self):
        super().updateWavenumberRange()

        wmin = min(self.data.wmin, octavvs.ui.constants.WMIN)
        wmax = max(self.data.wmax, octavvs.ui.constants.WMAX)
        self.lineEditMinwn.setRange(wmin, wmax, default=wmin)
        self.lineEditMaxwn.setRange(wmin, wmax, default=wmax)
        self.lineEditNormWavenum.setRange(wmin, wmax, default=.5*(wmin+wmax))

        if self.data.wavenumber is not None:
            # Update slider ranges before setting new values
            uitools.box_to_slider(
                    self.horizontalSliderMin, self.lineEditMinwn,
                    self.data.wavenumber, uitools.ixfinder_noless)
            uitools.box_to_slider(
                    self.horizontalSliderMax, self.lineEditMaxwn,
                    self.data.wavenumber, uitools.ixfinder_nomore)


    def loadFolder(self, *argv, **kwargs):
        assert self.rmiescRunning == 0
        return super().loadFolder(*argv, **kwargs)

    @pyqtSlot(int)
    def updateFile(self, num):
        super().updateFile(num)

        self.ptirMode = self.data.filetype == 'ptir'
        self.widgetSC.setHidden(self.ptirMode)
        self.widgetSCOptions.setHidden(self.ptirMode)
        self.plot_SC.setHidden(self.ptirMode)
        self.widgetMC.setHidden(not self.ptirMode)
        self.plot_MC.setHidden(not self.ptirMode)

        if self.bcNext:
            self.abcWorker.haltBC = True
            self.bcNext = True
        if self.acNext:
            self.abcWorker.haltBC = True
            self.acNext = True
        self.clearSC()

        self.updateWavenumberRange()
        self.updatedFile()


    @pyqtSlot(str, str, str)
    def showLoadErrorMessage(self, file, err, details):
        "Error message from loading a file in the worker thread, with abort option"
        q = self.loadErrorBox(file, (err, details if details else None))
        q.addButton('Skip file', QMessageBox.AcceptRole)
        q.addButton('Abort', QMessageBox.AcceptRole)
        if q.exec():
            self.scStop()


    # MC, mIRage correction
    def updateMC(self):
        wn = self.imageVisualizer.plot_spectra.getWavenumbers()
        indata = self.imageVisualizer.plot_spectra.getSpectra()
        corr = None
        if self.ptirMode and self.checkBoxMC.isChecked() and indata is not None:
            corr = ptir.normalize_mirage(
                wn, indata,
                endpoints=self.spinBoxMCEndpoints.value(),
                slopefactor=self.lineEditMCSlopefactor.value())[0]
        self.plot_MC.setData(wn, indata, corr)

    # AC, Atmospheric correction
    def loadACReference(self):
        startdir = resource_filename('octavvs', "reference_spectra")
        ref = self.getLoadFileName(
            "Load atmospheric reference spectrum",
            filter="Matrix file (*.mat)",
            settingname='atmRefDir', settingdefault=startdir)
        if not ref:
            return
        self.lineEditACReference.setText(ref)
        self.updateAC()

    def updateAC(self):
        wn = self.plot_MC.getWavenumbers()
        if wn is None:
            return
        indata = self.plot_MC.getSpectra()
        if not self.checkBoxAC.isChecked() or len(indata) == 0:
            self.plot_AC.setData(wn, indata, None)
            self.labelACInfo.setText('')
            return
        opt = dict(cut_co2=self.checkBoxSpline.isChecked(),
                   extra=self.checkBoxLocalPeak.isChecked(),
                   smooth=self.checkBoxSmoothCorrected.isChecked(),
                   ref=self.lineEditACReference.text())
        args = [ wn, indata, opt ]
        if self.acNext:
            self.acNext = args
        else:
            self.startAC.emit(*args)
            self.acNext = True

    @pyqtSlot(np.ndarray, np.ndarray, np.ndarray, np.ndarray)
    def acDone(self, wavenumber, indata, corr, factors):
        self.plot_AC.setData(wavenumber, indata, corr)
        if self.acNext and self.acNext is not True:
            self.startAC.emit(*self.acNext)
            self.acNext = True
        else:
            self.acNext = None
        if factors is not None:
            h = 'H<sub>2</sub>O: %.2f%%' % (factors[0] * 100)
            if factors[0] > .05:
                h = '<span style="color:red">%s</span>' % (h)
            c = 'CO<sub>2</sub>: %.2f%%' % (factors[1] * 100)
            self.labelACInfo.setText('Mean correction: %s %s' % (h, c))
        else:
            self.labelACInfo.setText('')

    @pyqtSlot(str)
    def acFailed(self, err):
        if err != '':
            self.errorMsg.showMessage('AC failed:<pre>\n' + err + '</pre>')
        self.acDone(None, None, None, None)


    # SC, Scattering correction
    def loadOtherReference(self):
        startdir = resource_filename('octavvs', "reference_spectra")
        ref = self.getLoadFileName(
            "Load reference spectrum for CRMieSC",
            filter="Matrix file (*.mat)",
            settingname='scRefDir', settingdefault=startdir)
        self.lineEditReferenceName.setText(ref)
        self.comboBoxReference.setCurrentText('Other')

    def scClustersToggle(self):
        self.spinBoxNclusScat.setEnabled(self.checkBoxClusters.isChecked())
        self.checkBoxStabilize.setEnabled(self.checkBoxClusters.isChecked())
        self.updateSC()

    def getSCSettings(self):
        """ Returns SC settings as a dict, to enable comparison with latest run.
        """
        params = self.getParameters()
        p = params.filtered('sc')
        # Make shadowed options irrelevant to the comparison
        if p['scRef'] != 'Other':
            p['scOtherRef'] = ''
        if p['scRef'] != 'Percentile':
            p['scRefPercentile'] = 0
        if not p['scClustering']:
            p['scClusters'] = 0
            p['scStable'] = False
        if p['scPCADynamic']:
            p['scPCA'] = 0
        else:
            p['scPCAMax'] = 0
            p['scPCAVariance'] = 0
        if not p['scAutoIters']:
            p['scMinImprov'] = 0
        # Include atm correction only if used
        if params.acDo:
            p.update(params.filtered('ac'))
        else:
            p['acDo'] = False
        # Assume that SC precludes MC
        p['mcDo'] = False
        # Selection of pixels to correct, or all if using clustering.
        p['selected'] = self.imageVisualizer.plot_raw.getSelected() if \
            not p['scClustering'] else None
        return p

    def scSettingsChanged(self, settings=None):
        """ Are these settings/data different from the previous ones used for SC?
            settings=None reads settings from the UI
        """
        if self.plot_SC.getWavenumbers() is None:
            return True
        if not self.scSettings:
            return True
        if not settings:
            settings = self.getSCSettings()
        return settings != self.scSettings

    def toggleSCPCA(self):
#        fixed = self.dialogSCAdvanced.radioButtonSCPCAFixed.isChecked()
#        self.dialogSCAdvanced.spinBoxSCPCA.setEnabled(fixed)
#        self.dialogSCAdvanced.spinBoxSCPCAMax.setEnabled(not fixed)
#        self.dialogSCAdvanced.lineEditSCPCAVariance.setEnabled(not fixed)
        self.updateSC()

    def updateSC(self):
        isperc = self.comboBoxReference.currentIndex() < 2
        self.stackedSC.setCurrentIndex(0 if isperc else 1)
        if self.rmiescRunning:
            return
        changed = self.scSettingsChanged()
        self.pushButtonSCRefresh.setEnabled(changed)

    def updateSCplot(self):
        self.plot_SC.setSelected(self.imageVisualizer.plot_raw.getSelected())

    def toggleRunning(self, newstate):
        onoff = not newstate
        self.fileLoader.setEnabled(onoff)
        self.lineEditSaveExt.setEnabled(onoff)
        self.comboBoxSaveFormat.setEnabled(onoff)
        self.pushButtonRun.setEnabled(onoff)
        self.pushButtonSCRefresh.setEnabled(onoff)
        self.dialogCreateReference.pushButtonCreateReference.setEnabled(onoff)
        if newstate:
            self.progressBarSC.setValue(0)
        self.pushButtonSCStop.setEnabled(newstate == 1)
        self.pushButtonRunStop.setEnabled(newstate == 2)
        self.dialogCreateReference.pushButtonStop.setEnabled(newstate == 3)
        self.rmiescRunning = newstate

    def refreshSC(self):
        if self.rmiescRunning:
            return
        self.scNewSettings = self.getSCSettings()
        self.toggleRunning(1)
        self.startRmiesc.emit(self.data, self.scNewSettings)


    def clearSC(self):
        if self.rmiescRunning == 1:
            return
        self.plot_SC.setData(self.plot_AC.getWavenumbers(),
                             None, None, None, None)
        self.scSettings = None

    def scStop(self):
        self.worker.halt = True
        self.pushButtonSCStop.setEnabled(False)
        self.pushButtonRunStop.setEnabled(False)
        self.dialogCreateReference.pushButtonStop.setEnabled(False)

    @pyqtSlot()
    def scStopped(self):
        self.plot_SC.endProgress()
        self.worker.halt = False
        if self.rmiescRunning == 1:
            self.scNewSettings = None
            self.toggleRunning(0)
        self.updateSC()

    @pyqtSlot(str, str)
    def scFailed(self, err, trace):
        self.errorMsg.showMessage(
            'Processing failed:<pre>\n' + err + "\n\n" + trace + '</pre>')
        self.scStopped()

    @pyqtSlot(np.ndarray, np.ndarray, np.ndarray)
    def scDone(self, wavenumber, indata, corr):
        self.scSettings = self.scNewSettings
        self.plot_SC.setData(wavenumber, indata, corr,
                             self.scSettings['selected'],
                             self.imageVisualizer.plot_raw.getSelected())
        self.scStopped()

    @pyqtSlot(int, int)
    def scProgress(self, a, b):
        if a < 0:
            self.progressBarSC.setFormat(
                ['', 'Atm %p%', 'RMieSC %p%', 'SGF %p%', 'Baseline',
                 'mIRage %p%'][-a])
        if a == -2:
            self.plot_SC.prepareProgress(self.plot_AC.getWavenumbers())
        self.progressBarSC.setValue(max(a, 0))
        self.progressBarSC.setMaximum(b)

    # SGF, Denoising with Savtisky-Golay filter
    def updateSGF(self):
        window = self.spinBoxWindowLength.value()
        if not window & 1:
            window += 1
            self.spinBoxWindowLength.setValue(window)
        order = self.spinBoxPolyOrder.value()
        if not self.ptirMode and self.checkBoxSC.isChecked():
            wn = self.plot_SC.getWavenumbers()
            indata = self.plot_SC.getSpectra()
        else:
            wn = self.plot_AC.getWavenumbers()
            indata = self.plot_AC.getSpectra()
        if (self.checkBoxSGF.isChecked() and indata is not None and
            len(indata) > 0 and window > order and indata.shape[1] >= window):
            corr = scipy.signal.savgol_filter(indata, window, order, axis=1)
        else:
            corr = None
        self.plot_SGF.setData(wn, indata, corr)


    # SR, Spectral region cutting
    def updateSR(self):
        wn = self.plot_SGF.getWavenumbers()
        indata = self.plot_SGF.getSpectra()
        if indata is None or not self.checkBoxSR.isChecked():
            self.plot_SR.setData(wn, indata, None, None, 1234, 2345)
            return
        cut = range(len(wn)-1 - self.horizontalSliderMax.value(),
                    len(wn) - self.horizontalSliderMin.value())
        if len(cut) < 2:
            self.plot_SR.setData(wn, indata, None, None, 1234, 2345)
        else:
            self.plot_SR.setData(wn, indata,
                                 self.plot_SGF.getSpectra()[:, cut],
                                 self.plot_SGF.getWavenumbers()[cut],
                                 self.lineEditMinwn.value(), self.lineEditMaxwn.value())

    def srMinSlide(self):
        uitools.slider_to_box(
                self.horizontalSliderMin, self.lineEditMinwn,
                self.data.wavenumber, uitools.ixfinder_noless)
        self.updateSR()

    def srMinEdit(self):
        uitools.box_to_slider(
                self.horizontalSliderMin, self.lineEditMinwn,
                self.data.wavenumber, uitools.ixfinder_noless)
        self.updateSR()

    def srMaxSlide(self):
        uitools.slider_to_box(
                self.horizontalSliderMax, self.lineEditMaxwn,
                self.data.wavenumber, uitools.ixfinder_nomore)
        self.updateSR()

    def srMaxEdit(self):
        uitools.box_to_slider(
                self.horizontalSliderMax, self.lineEditMaxwn,
                self.data.wavenumber, uitools.ixfinder_nomore)
        self.updateSR()

    # BC, Baseline Correction
    bcNames = ['rubberband', 'concaverubberband', 'asls', 'arpls', 'assymtruncq' ]
    def bcName(self):
        if not self.checkBoxBC.isChecked():
            return 'none'
        return self.bcNames[self.comboBoxBaseline.currentIndex()]

    def bcSetMethod(self, val):
        if val in self.bcNames:
            self.checkBoxBC.setChecked(True)
            self.comboBoxBaseline.setCurrentIndex(self.bcNames.index(val))
        else:
            self.checkBoxBC.setChecked(False)
            self.comboBoxBaseline.setCurrentIndex(0)

    def updateBC(self):
        indata = self.plot_SR.getSpectra()
        wn = self.plot_SR.getWavenumbers()
        bc = self.bcName()
        if indata is None or bc == 'none':
            self.plot_BC.setData(wn, indata, None)
            return
        if bc == 'asls':
            param = {'lam': self.lineEditLambda.value(),
                     'p': self.lineEditP.value()}
        elif bc == 'arpls':
            param = {'lam': self.lineEditLambdaArpls.value()}
        elif bc == 'rubberband':
            param = {}
        elif bc == 'concaverubberband':
            param = {'iters': self.spinBoxItersBC.value()}
        elif bc == 'assymtruncq':
            param = {'poly': self.spinBoxBCPoly.value(),
                     'thresh': self.lineEditBCThresh.value()}
        if self.bcNext:
            self.abcWorker.haltBC = True
            self.bcNext = [ wn, indata, bc, param ]
        else:
            self.startBC.emit(wn, indata, bc, param)
            self.bcNext = True

    @pyqtSlot(np.ndarray, np.ndarray, np.ndarray)
    def bcDone(self, wavenumber, indata, corr):
        if(wavenumber is not None):
            self.plot_BC.setData(wavenumber, indata, corr)
        self.abcWorker.haltBC = False
        if not self.bcNext or self.bcNext is True:
            self.bcNext = None
        else:
            self.startBC.emit(*self.bcNext)
            self.bcNext = True

    @pyqtSlot(str)
    def bcFailed(self, err):
        if err != '':
            print(err)
            self.errorMsg.showMessage('BC failed:<pre>\n' + err + '</pre>')
        self.bcDone(None, None, None)

    def bcMethod(self):
        self.bcParams.setCurrentIndex(self.comboBoxBaseline.currentIndex())
        self.updateBC()

    def connectLogSliders(self):
        def expSliderPos(box, slider, rng):
            return round(slider.maximum() * (np.log10(box.value()) - rng[0]) / (rng[1] - rng[0]))

        def expBoxVal(slider, rng):
            return 10.**(slider.value() / slider.maximum() * (rng[1] - rng[0]) + rng[0])

        def makeSliderFuncs(box, slider, range_):
            def slide():
                if (not box.hasAcceptableInput() or
                    expSliderPos(box, slider, range_) != slider.value()):
                        box.setValue(expBoxVal(slider, range_))
                self.updateBC()
            def edit():
                if box.hasAcceptableInput():
                    slider.setValue(expSliderPos(box, slider, range_))
                else:
                    box.setValue(expBoxVal(slider, range_))
                self.updateBC()
            box.editingFinished.connect(edit)
            slider.valueChanged.connect(slide)
            box.setRange(*(10. ** range_))
            box.setFormat("%.4g")
            edit()
        makeSliderFuncs(self.lineEditLambda,
                        self.horizontalSliderLambda, self.bcLambdaRange)
        makeSliderFuncs(self.lineEditLambdaArpls,
                        self.horizontalSliderLambdaArpls, self.bcLambdaRange)
        makeSliderFuncs(self.lineEditP,
                        self.horizontalSliderP, self.bcPRange)


    # Normalization
    normNames = [ 'mean', 'area', 'wn', 'max', 'n2' ]
    def normName(self):
        return self.normNames[self.comboBoxNormMethod.currentIndex()]
    def normIndex(self, val):
        if val not in self.normNames:
            return 0
        return self.normNames.index(val)

    def updateNorm(self):
        meth = self.normName()
        self.lineEditNormWavenum.setEnabled(meth == 'wn')
        wn = self.plot_BC.getWavenumbers()
        indata = self.plot_BC.getSpectra()
        if indata is None or len(indata) == 0 or not self.checkBoxNorm.isChecked():
            self.plot_norm.setData(wn, indata, indata)
            return
        if not self.lineEditNormWavenum.hasAcceptableInput():
            self.lineEditNormWavenum.setValue()

        n = normalization.normalize_spectra(
                meth, y=indata, wn=wn, wavenum=self.lineEditNormWavenum.value())
        self.plot_norm.setData(wn, indata, n)


    # Loading and saving parameters
    def getParameters(self):
        "Copy from UI to some kind of parameters object"
        p = PrepParameters()
        self.fileLoader.saveParameters(p)
        self.imageVisualizer.saveParameters(p)

        p.saveExt = self.lineEditSaveExt.text()
        p.saveFormat = self.comboBoxSaveFormat.currentText()
        p.acDo = self.checkBoxAC.isChecked()
        p.acSpline = self.checkBoxSpline.isChecked()
        p.acLocal = self.checkBoxLocalPeak.isChecked()
        p.acSmooth = self.checkBoxSmoothCorrected.isChecked()
        p.acReference = self.lineEditACReference.text()
        p.scDo = self.checkBoxSC.isChecked()
        p.scRef = self.comboBoxReference.currentText()
        p.scOtherRef = self.lineEditReferenceName.text()
        p.scRefPercentile = self.lineEditSCRefPercentile.value()
        p.scIters = self.spinBoxNIteration.value()
        p.scClustering = self.checkBoxClusters.isChecked()
        p.scClusters = self.spinBoxNclusScat.value()
        p.scStable = self.checkBoxStabilize.isChecked()
        p.setAlgorithm(self.dialogSCAdvanced.comboBoxSCAlgo.currentText())
        p.scResolution = self.dialogSCAdvanced.spinBoxSCResolution.value()
        p.scAmin = self.dialogSCAdvanced.lineEditSCamin.value()
        p.scAmax = self.dialogSCAdvanced.lineEditSCamax.value()
        p.scDmin = self.dialogSCAdvanced.lineEditSCdmin.value()
        p.scDmax = self.dialogSCAdvanced.lineEditSCdmax.value()
        p.scConstant = self.dialogSCAdvanced.checkBoxSCConstant.isChecked()
        p.scLinear = self.dialogSCAdvanced.checkBoxSCLinear.isChecked()
        p.scPrefitReference = self.dialogSCAdvanced.checkBoxSCPrefitReference.isChecked()
        p.scPenalize = self.dialogSCAdvanced.checkBoxSCPenalize.isChecked()
        p.scPenalizeLambda = self.dialogSCAdvanced.lineEditSCPenalizeLambda.value()
        p.scPCADynamic = self.dialogSCAdvanced.radioButtonSCPCADynamic.isChecked()
        p.scPCA = self.dialogSCAdvanced.spinBoxSCPCA.value()
        p.scPCAMax = self.dialogSCAdvanced.spinBoxSCPCAMax.value()
        p.scPCAVariance = self.dialogSCAdvanced.lineEditSCPCAVariance.value()
        p.scAutoIters = self.dialogSCAdvanced.checkBoxSCAutoIters.isChecked()
        p.scMinImprov = self.dialogSCAdvanced.lineEditSCMinImprov.value()
        p.mcDo = self.checkBoxMC.isChecked()
        p.mcEndpoints = self.spinBoxMCEndpoints.value()
        p.mcSlopefactor = self.lineEditMCSlopefactor.value()
        p.sgfDo = self.checkBoxSGF.isChecked()
        p.sgfWindow = self.spinBoxWindowLength.value()
        p.sgfOrder = self.spinBoxPolyOrder.value()
        p.srDo = self.checkBoxSR.isChecked()
        p.srMin = self.lineEditMinwn.value()
        p.srMax = self.lineEditMaxwn.value()
        p.bcDo = self.checkBoxBC.isChecked()
        p.bcMethod = self.bcName()
        p.bcIters = self.spinBoxItersBC.value()
        p.bcLambda = self.lineEditLambda.value()
        p.bcLambdaArpls = self.lineEditLambdaArpls.value()
        p.bcP = self.lineEditP.value()
        p.bcThreshold = self.lineEditBCThresh.value()
        p.bcPoly = self.spinBoxBCPoly.value()
        p.normDo = self.checkBoxNorm.isChecked()
        p.normMethod = self.normName()
        p.normWavenum = self.lineEditNormWavenum.value()
        return p

    def saveParameters(self):
        filename = self.getSaveFileName(
                "Save preprocessing settings",
                filter="Setting files (*.pjs);;All files (*)",
                settingname='settingsDir',
                suffix='pjs')
        if filename:
            try:
                self.getParameters().save(filename)
            except Exception as e:
                self.showDetailedErrorMessage(
                        "Error saving settings to "+filename+": "+repr(e),
                        traceback.format_exc())

    def setParameters(self, p):
        "Copy from some kind of parameters object to UI"
        self.spinBoxSpectra.setValue(0)
        self.fileLoader.loadParameters(p)
        self.imageVisualizer.loadParameters(p)

        self.lineEditSaveExt.setText(p.saveExt)
        self.comboBoxSaveFormat.setCurrentText(p.saveFormat)
        self.checkBoxAC.setChecked(p.acDo)
        self.checkBoxSpline.setChecked(p.acSpline)
        self.checkBoxLocalPeak.setChecked(p.acLocal)
        self.checkBoxSmoothCorrected.setChecked(p.acSmooth)
        self.lineEditACReference.setText(p.acReference)
        self.checkBoxSC.setChecked(p.scDo)
        self.comboBoxReference.setCurrentText(p.scRef)
        self.lineEditReferenceName.setText(p.scOtherRef)
        self.lineEditSCRefPercentile.setValue(p.scRefPercentile)
        self.spinBoxNIteration.setValue(p.scIters)
        self.spinBoxNclusScat.setValue(p.scClusters)
        self.checkBoxClusters.setChecked(p.scClustering)
        self.checkBoxStabilize.setChecked(p.scStable)

        # Find the right algorithm - the name just needs to be in there somewhere
        cb = self.dialogSCAdvanced.comboBoxSCAlgo
        algnum = next((i for i in range(cb.count()) if
                       p.scAlgorithm in cb.itemText(i).lower()), None)
        if algnum is not None:
            cb.setCurrentIndex(algnum)

        self.dialogSCAdvanced.spinBoxSCResolution.setValue(p.scResolution)
        self.dialogSCAdvanced.lineEditSCamin.setValue(p.scAmin)
        self.dialogSCAdvanced.lineEditSCamax.setValue(p.scAmax)
        self.dialogSCAdvanced.lineEditSCdmin.setValue(p.scDmin)
        self.dialogSCAdvanced.lineEditSCdmax.setValue(p.scDmax)
        self.dialogSCAdvanced.checkBoxSCConstant.setChecked(p.scConstant)
        self.dialogSCAdvanced.checkBoxSCLinear.setChecked(p.scLinear)
        self.dialogSCAdvanced.checkBoxSCPrefitReference.setChecked(p.scPrefitReference)
        self.dialogSCAdvanced.radioButtonSCPCADynamic.setChecked(p.scPCADynamic)
        self.dialogSCAdvanced.spinBoxSCPCA.setValue(p.scPCA)
        self.dialogSCAdvanced.spinBoxSCPCAMax.setValue(p.scPCAMax)
        self.dialogSCAdvanced.lineEditSCPCAVariance.setValue(p.scPCAVariance)
        self.dialogSCAdvanced.checkBoxSCAutoIters.setChecked(p.scAutoIters)
        self.dialogSCAdvanced.lineEditSCMinImprov.setValue(p.scMinImprov)
        self.checkBoxMC.setChecked(p.mcDo)
        self.spinBoxMCEndpoints.setValue(p.mcEndpoints)
        self.lineEditMCSlopefactor.setValue(p.mcSlopefactor)

        self.checkBoxSGF.setChecked(p.sgfDo)
        self.spinBoxWindowLength.setValue(p.sgfWindow)
        self.spinBoxPolyOrder.setValue(p.sgfOrder)
        self.checkBoxSR.setChecked(p.srDo)
        self.lineEditMinwn.setValue(p.srMin)
        self.lineEditMaxwn.setValue(p.srMax)
        self.checkBoxBC.setChecked(p.bcDo)
        self.bcSetMethod(p.bcMethod)
        self.spinBoxItersBC.setValue(p.bcIters)
        self.lineEditLambda.setValue(p.bcLambda)
        self.lineEditLambda.editingFinished.emit()
        self.lineEditLambdaArpls.setValue(p.bcLambdaArpls)
        self.lineEditLambdaArpls.editingFinished.emit()
        self.lineEditP.setValue(p.bcP)
        self.lineEditP.editingFinished.emit()
        self.lineEditBCThresh.setValue(p.bcThreshold)
        self.spinBoxBCPoly.setValue(p.bcPoly)
        self.checkBoxNorm.setChecked(p.normDo)
        self.comboBoxNormMethod.setCurrentIndex(self.normIndex(p.normMethod))
        self.lineEditNormWavenum.setValue(p.normWavenum)

        self.srMinEdit()
        self.srMaxEdit()
        self.wavenumberEdit()

        self.updateSGF()

    def loadParameters(self, checked=False, filename=None):
        """
        Load parameters from a file, showing a dialog if no filename is given.
        The 'checked' parameter is just a placeholder to match QPushButton.clicked().
        """
        if not filename:
            filename = self.getLoadFileName(
                    "Load preprocessing settings",
                    filter="Settings files (*.pjs);;All files (*)",
                    settingname='settingsDir')
        if filename:
            p = self.getParameters()
            try:
                p.load(filename)
                self.setParameters(p)
            except Exception as e:
                self.showDetailedErrorMessage(
                        "Error loading settings from "+filename+": "+repr(e),
                        traceback.format_exc())


    def runBatch(self, checked=False, foldername=None):
        """
        Run the selected preprocessing steps, showing a dialog if no output path is given.
        The 'checked' parameter is just a placeholder to match QPushButton.clicked().
        """
        assert not self.rmiescRunning
        if len(self.data.filenames) < 1:
            self.errorMsg.showMessage('Load some data first')
            return
        params = self.getParameters()
        if self.ptirMode:
            params.scDo = False
        else:
            params.mcDo = False

        if foldername is None:
            foldername = self.getDirectoryName("Select save directory",
                                               settingname='spectraDir',
                                               savesetting=False)
        if not foldername:
            return
        preservepath = False

        all_paths = { os.path.dirname(f) for f in self.data.filenames }
        folder_within_input = any(p.startswith(foldername) for p in all_paths)

        if folder_within_input:
            if params.saveExt == '':
                self.errorMsg.showMessage('A filename save suffix must be specified '+
                                          'when loading and saving files in the same directory')
                return
            elif len(self.data.filenames) > 1:
                yn = QMessageBox.question(self, 'Identical directory',
                                          'Are you sure you want to '+
                                          'save output files in the input directory')
                if yn != QMessageBox.Yes:
                    return
            preservepath = True
        elif len(all_paths) > 1:
            yn = QMessageBox.question(self, 'Multiple directories',
                      'Input files are in multiple directories. Should this '+
                      'directory structure be replicated in the output directory?')
            preservepath = yn == QMessageBox.Yes

        self.toggleRunning(2)
        self.scNewSettings = self.getSCSettings()
        self.startBatch.emit(self.data, params, foldername, preservepath)

    def createReference(self, checked=False):
        """
        Run the selected preprocessing steps (sans SC and SR) on all data, then
        save the average spectrum as a reference for scattering correction.
        The 'checked' parameter is just a placeholder to match QPushButton.clicked().
        """
        assert not self.rmiescRunning
        if len(self.data.filenames) < 1:
            self.errorMsg.showMessage('Load some data first')
            return
        params = self.getParameters()

        startdir = resource_filename('octavvs', "reference_spectra")
        filename = self.getLoadFileName(
            "Save reference spectrum",
            savesuffix='mat',
            filter="Matrix file (*.mat)",
            settingname='scRefDir', settingdefault=startdir)
        if not filename:
            return

        self.toggleRunning(3)
        self.scNewSettings = self.getSCSettings()
        self.startCreateReference.emit(self.data, params, filename)
        self.lineEditReferenceName.setText(filename)

    @pyqtSlot(int, int)
    def batchProgress(self, a, b):
        self.progressBarRun.setValue(a)
        self.progressBarRun.setMaximum(b)

    @pyqtSlot(bool)
    def batchDone(self, success):
        self.worker.halt = False
        self.progressBarSC.setValue(0)
        self.progressBarRun.setValue(0)
        # if err != '':
        #     self.errorMsg.showMessage(err)
        if success and self.rmiescRunning == 3:
            self.comboBoxReference.setCurrentText('Other')
        self.toggleRunning(0)


def main():
    parser = argparse.ArgumentParser(
            description='Graphical application for preprocessing of hyperspectral data.')
    parser.add_argument('files', metavar='file', nargs='*',
                        help='initial hyperspectral images to load')
    parser.add_argument('-p', '--params', metavar='file.pjs', dest='paramFile',
                        help='parameter file to load')
    parser.add_argument('-r', '--run', metavar='output_dir', nargs='?',
                        dest='savePath', const='./',
                        help='runs and saves to the output directory (params '+
                        'argument must also be passed)')
    MyMainWindow.run_octavvs_application(parser=parser,
                                         parameters=['files', 'paramFile', 'savePath'])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 17:25:10 2020

@author: carl
"""

import numpy as np

from . import constants

def ixfinder_nearest(wn, val):
    return len(wn)-1 - (np.abs(wn - val)).argmin()

def ixfinder_noless(wn, val):
    return wn[::-1].searchsorted(val, 'left')

def ixfinder_nomore(wn, val):
    return wn[::-1].searchsorted(val, 'right')-1

def slider_to_box(slider, box, wn, ixfinder):
    """Set a QLineEdit from the value on a slider, discretized by
    the wavenumber array wn"""
    if wn is not None:
        if (not box.hasAcceptableInput() or
            slider.value() != ixfinder(wn, box.value())):
            box.setValue(wn[-1-slider.value()])
    elif (not box.hasAcceptableInput() or
            slider.value() != int(round(slider.maximum() *
            (box.value() - constants.WMIN) / (constants.WMAX - constants.WMIN)))):
        box.setValue(constants.WMIN + (constants.WMAX - constants.WMIN) *
                     slider.value() / slider.maximum())

def box_to_slider(slider, box, wn, ixfinder):
    """Set a slider from the value in a lineedit, discretized by
    the wavenumber array wn"""
    oldb = slider.blockSignals(True)
    if wn is not None:
        # assert slider.maximum() == len(wn) - 1
        if slider.maximum() != len(wn) - 1:
            slider.setMaximum(len(wn) - 1)
        if box.hasAcceptableInput():
            slider.setValue(ixfinder(wn, box.value()))
        else:
            box.setValue(wn[-1-slider.value()])
    else:
        slider.setValue(int(round(slider.maximum() *
              (box.value() - constants.WMIN) / (constants.WMAX - constants.WMIN))))
    slider.blockSignals(oldb)

